from functools import partial

from flask import g, request
from flask_holster.exts import ext_dict, guess_type
from flask_holster.mime import Accept
from flask_holster.parts import bare_holster, holsterize
from flask_holster.views import HTMLTemplate, templates


def with_template(mime, templater):
    """
    Decorator to add a customized templater to a holstered view.

    This decoration must happen prior to the holstering:

        @app.holster("/example")
        @with_template("html", HTMLTemplater("example.html"))
        def example():
            return {}
    """

    # Extensions are allowed too!
    if mime in ext_dict:
        mime = ext_dict[mime]

    def attach(f):
        d = getattr(f, "_holsters", {})
        d[mime] = templater
        f._holsters = d
        return f
    return attach


def html_template(template):
    return with_template("text/html", HTMLTemplate(template=template,
        from_string=False))


def holster(app, route, **kwargs):
    """
    Decorator which replaces ``route()``.

    This decorator is nested because it is meant to be called in the same
    style as ``route()``:

        @app.holster("/")
        def index():
            pass
    """

    def inner(view):
        wrapped = holsterize(view)
        # This returns the wrapped view. We don't care about it though.
        bare_holster(app, route, **kwargs)(wrapped)

        # Return the original view so that people can do more things with it.
        # Even if they re-holster the view, it's gonna be way easier for us to
        # do things if we don't multiply-wrap it.
        return view

    return inner


def holster_url_value_preprocessor(endpoint, values):
    """
    Figure out which MIME type should be returned, and prepare the value for
    later processing.
    """

    # Check to see if this has already happened, e.g. in a blueprint.
    if hasattr(g, "_holster_mime"):
        return

    types = Accept(",".join(templates))
    if values and "ext" in values:
        ext = values.pop("ext")
        types.prefer(guess_type(ext))

    accept = Accept(request.headers["accept"])

    mime = accept.best(types)

    g._holster_mime = mime


def init_holster(app):
    """
    Initialize a Flask application or blueprint to have holsters.

    This is mostly just attaching hooks and setting default configuration
    values.
    """

    app.bare_holster = partial(bare_holster, app)
    app.holster = partial(holster, app)
    app.holsterize = holsterize
    app.url_value_preprocessor(holster_url_value_preprocessor)

    if hasattr(app, "config"):
        app.config.setdefault("HOLSTER_COMPRESS", False)



from functools import wraps

def lift(f):
    """
    "Decoratorial" lift.

    This is a functorial map specialized over functions, and I could talk
    about the category theory behind it, but the main takeaway is that you can
    use this as a decorator to lift a function on data to a function between
    functions on that data.

    >>> def first(x):
    ...  return x + 1
    ...
    >>> def second(x):
    ...  return x * 2
    ...
    >>> def third(x):
    ...  return x ** 3
    ...
    >>> @lift(third)
    ... @lift(second)
    ... @lift(first)
    ... def const():
    ...  return 1
    ...
    >>> const()
    64
    """

    @wraps(f)
    def deco(g):
        @wraps(g)
        def inner(*args, **kwargs):
            return f(g(*args, **kwargs))
        return inner
    return deco



from functools import partial, wraps
from zlib import compress

from flask import current_app, g, make_response, request
from flask_holster.views import Str, templates


def _worker(view, *args, **kwargs):
    """
    The actual worker.
    """

    d = view(*args, **kwargs)

    if callable(d):
        # We've been hoodwinked! It's pretty likely that d is a handcrafted
        # response. Should we let it through as-is? Yes.
        return d

    mime = g._holster_mime.plain()

    overrides = getattr(view, "_holsters", {})
    templater = overrides.get(mime)
    if not templater:
        templater = templates.get(mime, Str())

    # Run the templater.
    data = templater.format(d)

    # Optionally compress the data.
    compress_setting = current_app.config.get("HOLSTER_COMPRESS", False)
    if compress_setting and "deflate" in request.accept_encodings:
        # It's possible that our data is Unicode. Thanks, Jinja. In that case,
        # turn it into UTF-8 before compressing.
        if isinstance(data, unicode):
            data = data.encode("utf-8")
        data = compress(data)
        compressed = True
    else:
        compressed = False

    # Use the formatted data as the response body.
    response = make_response(data)

    # Indicate the MIME type of the data that we are sending back.
    response.headers["Content-Type"] = mime

    # Indicate to caches that the data returned is dependent on the Accept
    # header's value in the request.
    response.vary.add("Accept")

    # Fill out the compression header and indicate varyings, if needed.
    if compressed:
        response.headers["Content-Encoding"] = "deflate"
        response.vary.add("Accept-Encoding")

    return response


def holsterize(view):
    """
    Wrap a view with a holster.
    """

    p = wraps(view)(partial(_worker, view))

    return p


def extend(route):
    """
    Add a file extension to a route.

    This function adds a file extension, called "ext", to a route.

    If the route ends with a trailing slash, then the extension precedes the
    slash. This preserves the standard directory-redirect mechanism of
    Werkzeug.

    If the route is the root, then the extension is tacked onto the end. This
    looks awkward but there's no better solution.
    """

    if route == "/":
        return "/.<ext>"
    elif route.endswith("/"):
        return "%s.<ext>/" % route[:-1]
    else:
        return "%s.<ext>" % route


def bare_holster(app, route, **kwargs):
    """
    Decorator which replaces ``route()``.

    This decorator is nested because it is meant to be called in the same
    style as ``route()``:

        @app.bare_holster("/")
        def index():
            pass

    This is the core functionality of routing; it does not include the logic
    for holstering views, only for attaching them to the app.
    """

    extended = extend(route)

    def inner(view):
        name = view.__name__
        app.add_url_rule(route, endpoint=name, view_func=view, **kwargs)
        app.add_url_rule(extended, endpoint="%s-ext" % name, view_func=view,
                **kwargs)
        return view

    return inner



from flask import json, render_template, render_template_string

default_html_template = """
<meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
<meta charset="utf-8" />
{% macro dump(d) %}
    <ul>
    {% for k in d %}
        {% set v = d[k] %}
        {% if v is defined %}
            <li>{{ k }}:
                {% if v is mapping %}
                <ul>{{ dump(v) }}</ul>
                {% else %}
                {{ v|e }}
                {% endif %}
            </li>
        {% else %}
            <li>{{ k }}</li>
        {% endif %}
    {% endfor %}
    </ul>
{% endmacro %}
{{ dump(d) }}
"""


class HTMLTemplate(object):

    def __init__(self, template=None, from_string=True):
        self.template = template if template else default_html_template
        self.from_string = from_string

    def format(self, d):
        if self.from_string:
            return render_template_string(self.template, d=d)
        else:
            return render_template(self.template, d=d)



class JSONRenderer(object):

    def format(self, d):
        return json.dumps(d)



class Str(object):

    def format(self, d):
        return str(d)



templates = {
    "text/html":        HTMLTemplate(),
    "application/json": JSONRenderer(),
    "text/plain":       Str(),
}

# YAML. Requires the yaml module to be installed.
try:
    from yaml import safe_dump

    class YAMLRenderer(object):

        def format(self, d):
            return safe_dump(d)

    templates["application/x-yaml"] = YAMLRenderer()
except ImportError:
    pass



ext_dict = {
    "html": "text/html",
    "json": "application/json",
    # The YAML MIME type is not standardized. This is the same type Rails
    # uses; hopefully it is close enough.
    "yaml": "application/x-yaml",
}

def guess_type(ext):
    """
    Guess which MIME type fits the given extension.
    """

    return ext_dict.get(ext, "text/plain")



def match_with_wildcard(s1, s2):
    """
    If either parameter is the wildcard, match. Otherwise, check for equality.
    """

    if s1 == "*":
        return s2
    elif s2 == "*":
        return s1
    elif s1 == s2:
        return s1
    else:
        return None

class MIME(object):
    """
    A MIME type.
    """

    def __init__(self, category, filetype, q):
        self.category = category
        self.filetype = filetype
        self.params = {"q": q}

    def __str__(self):
        params = ", ".join("%s = %s" % t for t in self.params.iteritems())
        return "%s (%s)" % (self.plain(), params)

    __repr__ = __str__

    @classmethod
    def from_string(cls, s):
        pieces = s.split(";")
        category, filetype = pieces[0].split("/")
        self = cls(category, filetype, 1.0)

        for piece in pieces[1:]:
            if "=" in piece:
                k, v = piece.split("=", 1)
                # Special case for q.
                if k == "q":
                    v = float(v)
                self.params[k] = v

        return self

    def plain(self):
        return "%s/%s" % (self.category, self.filetype)

    def match(self, other):
        """
        Create a new MIME type representing the quality of the match between
        this type and the other type.
        """

        category = match_with_wildcard(self.category, other.category)
        filetype = match_with_wildcard(self.filetype, other.filetype)

        if category and filetype:
            return MIME(category, filetype, self.params["q"] *
                other.params["q"])
        else:
            return MIME("*", "*", 0.0)

class Accept(object):
    """
    The Accept header, as defined by RFC 2616.
    """

    def __init__(self, s):
        self.types = [MIME.from_string(x) for x in s.split(",")]

    def __str__(self):
        return "<Accept [%s]>" % ", ".join(str(x) for x in self.types)

    __repr__ = __str__

    def prefer(self, s):
        """
        Parse a MIME type and prefer it over the other types in this listing.
        """

        mime = MIME.from_string(s)
        for t in self.types:
            t.params["q"] *= 0.5
        self.types.append(mime)

    def match(self, other):
        """
        Generate a list of matched MIME types from another Accept listing.
        """

        # As long as we can't have itertools.product(), we have to do this the
        # slow way.
        l = []

        for x in self.types:
            for y in other.types:
                l.append(x.match(y))

        return l

    def best(self, other):
        """
        Given another Accept listing, choose the best acceptable MIME type.

        None will be returned if no acceptable type could be found.
        """

        l = self.match(other)
        l.sort(key=lambda mime: mime.params["q"], reverse=True)
        if l:
            return l[0]
        return None



#!/usr/bin/env python

from setuptools import setup

setup(
    name="Flask-Holster",
    version="0.2.5",
    url="http://github.com/MostAwesomeDude/flask-holster",
    license="MIT",
    author="Corbin Simpson",
    author_email="cds@corbinsimpson.com",
    description="Rigid MVC content negotiation for Flask",
    long_description=open("README.rst").read(),
    packages=["flask_holster"],
    test_suite="flask_holster.test",
    zip_safe=False,
    platforms="any",
    install_requires=[
        "Flask",
    ],
    classifiers=[
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



"""
Oftentimes, one simply wants to get a common task done. This module faciliates
that in the context of Holster.
"""

from flask_holster.main import with_template
from flask_holster.views import HTMLTemplate

def html(template):
    """
    Render this holstered handler with the given template.

    The template is supposed to be a file path; think of this helper as
    replacing ``render_template()``.
    """

    return with_template("text/html", HTMLTemplate(template,
        from_string=False))
